# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function, absolute_import

import logging
logger = logging.getLogger(__name__)

import os
import signal
import resource

from collections import namedtuple

from spyne import ComplexModel, Double, Integer32, UnsignedInteger32, M, Unicode
from spyne.util import memoize

RLimitEntry = namedtuple("RLimitEntry", "name unit type mult")


# Specifying stuff in bytes in config files could be inconvenient so we specify
# units and multipliers to make it easy to generate the config options
def _check_and_add_rlimit(name, unit, type, mult, **kwargs):
    aname = "RLIMIT_{}".format(name)
    if hasattr(resource, aname):
        RLIMIT_DICT[getattr(resource, aname)] = \
                                   RLimitEntry(name, unit, type(**kwargs), mult)


RLIMIT_DICT = {}


_check_and_add_rlimit("CORE", 'mb', Double, 2 ** 20,
    help="""The maximum size (in megabytes) of a core file that the current 
    process can create. This may result in the creation of a partial core file 
    if a larger core would be required to contain the entire process image.""")

_check_and_add_rlimit("CPU", 'h', Double, 3600,
    help="""The maximum amount of processor time (in hours) that a process can 
    use. If this limit is exceeded, a SIGXCPU signal is sent to the process.
    (See the signal module documentation for information about how to catch
    this signal and do something useful, e.g. flush open files to disk.)""")

_check_and_add_rlimit("FSIZE", 'mb', Double, 2 ** 20,
    help="""The maximum size (in bytes) of a file which the process may 
    create.""")

_check_and_add_rlimit("DATA", 'mb', Double, 2 ** 20,
    help="""The maximum size (in bytes) of the process’s heap.""")

_check_and_add_rlimit("STACK", 'mb', Double, 2 ** 20,
    help="""The maximum size (in bytes) of the call stack for the current 
    process. This only affects the stack of the main thread in a multi-threaded 
    process.""")

_check_and_add_rlimit("RSS", 'mb', Double, 2 ** 20,
    help="""The maximum resident set size that should be made available to the
    process.""")

_check_and_add_rlimit("NPROC", '', Integer32, 1,
    help="""The maximum number of processes the current process may create.""")

_check_and_add_rlimit("NOFILE", '', Integer32, 1,
    help="""The maximum number of open file descriptors for the current 
    process.""")

_check_and_add_rlimit("OFILE", '', Integer32, 1,
    help="""# The BSD name for RLIMIT_NOFILE.""")

_check_and_add_rlimit("MEMLOCK", 'mb', Double, 2 ** 20,
    help="""The maximum address space which may be locked in memory.""")

_check_and_add_rlimit("AS", 'mb', Double, 2 ** 20,
    help="""The maximum area (in bytes) of address space which may be taken by 
    the process.""")


class Limits(ComplexModel):
    _type_info = [
        ("{}_{}".format(v.name.lower(), v.unit), v.type(rlimit=k))
        if v.unit else
        (v.name.lower(), v.type(rlimit=k))

        for k, v in RLIMIT_DICT.items()
    ]

    def apply_limits_impl(self, limtype):
        for k, v in self.get_flat_type_info(self.__class__).items():
            val = getattr(self, k, None)
            if val is None:
                continue

            curr_rlimit = Limits._type_info[k].Attributes.rlimit
            oldlim = resource.getrlimit(curr_rlimit)

            newlim = list(oldlim)
            newlim[limtype] = int(val * 1024 ** 2)
            newlim = tuple(newlim)

            resource.setrlimit(curr_rlimit, newlim)
            logger.debug("RLIMIT_%s: was: %r now: %r", k, oldlim, newlim)


TIMED_ACTION_MAP = {
    'SIGTERM': lambda *_: os.kill(os.getpid(), signal.SIGTERM),
    'SIGKILL': lambda *_: os.kill(os.getpid(), signal.SIGKILL),
}


@memoize
def TTimedLimit(ValueType, rlimit):
    class TimedLimit(ComplexModel):
        __type_name__ = "Timed{}Limit".format(ValueType.get_type_name())

        _type_info = [
            ('value', M(ValueType)),
            ('action', M(Unicode(values=list(TIMED_ACTION_MAP.keys())))),
            ('period_s', UnsignedInteger32(default=60)),
            ('num_breaches', UnsignedInteger32(default=3)),
        ]

        def apply(self, name):
            logger.info("Enabling timed %s limit of %.1f %s "
                            "with period: %d seconds, num. breaches: %d",
                                  name, self.value, RLIMIT_DICT[rlimit].unit,
                                               self.period_s, self.num_breaches)
            return enforce_timed_limit(name, self, rlimit)

    return TimedLimit


class TimedLimits(ComplexModel):
    _type_info = [
        ("{}_{}".format(v.name.lower(), v.unit),
                      TTimedLimit(v.type, rlimit=k).customize(not_wrapped=True))

        if v.unit else

        (v.name.lower(),
                      TTimedLimit(v.type, rlimit=k).customize(not_wrapped=True))

        for k, v in RLIMIT_DICT.items()
    ]

    def apply(self):
        for k, v in self.get_flat_type_info(self.__class__).items():
            val = getattr(self, k, None)
            if val is None:
                continue

            val.apply(k)



class LimitsChoice(ComplexModel):
    _type_info = [
        ('soft', Limits.customize(not_wrapped=True)),
        ('hard', Limits.customize(not_wrapped=True)),
        ('timed', TimedLimits.customize(not_wrapped=True)),
    ]

    def apply(self):
        # Remember:
        #     soft, hard = resource.getrlimit(whatever)
        SOFT = 0
        HARD = 1

        if self.soft is not None:
            self.soft.apply(SOFT)

        if self.hard is not None:
            self.hard.apply(HARD)

        if self.timed is not None:
            self.timed.apply()


def enforce_timed_limit(name, limconf, rlimit):
    from twisted.internet.task import LoopingCall
    from twisted.internet.threads import deferToThread
    from neurons.daemon.config.daemon import meminfo

    status = [0]
    rlimit = RLIMIT_DICT[rlimit]
    assert isinstance(rlimit, RLimitEntry)
    callback = TIMED_ACTION_MAP[limconf.action]

    def _check():
        try:
            val = meminfo().rss

            if val < (limconf.value * rlimit.mult):
                if status[0] > 0:
                    logger.warning("%s limit went below treshold l: %.1f "
                              "c: %.1f", name, limconf.value, val / rlimit.mult)

                status[0] = 0
                return

            status[0] += 1

            if status[0] < limconf.num_breaches:
                logger.warning("%s limit reached, %d checks left "
                       "before taking %s action. s: %d l: %.1f c: %.1f",
                       name, (limconf.num_breaches - status[0]), limconf.action,
                                    status[0], limconf.value, val / rlimit.mult)
                return

            logger.warning("%s limit reached, performing %s. "
                           "s: %d l: %.1f c: %.1f", name, limconf.action,
                                    status[0], limconf.value, val / rlimit.mult)

            callback(limconf.value, limconf.period_s, limconf.num_breaches)

        except Exception as e:
            logger.error("Error in memory check job:")
            logger.exception(e)

    if meminfo is None:
        logger.warning("meminfo() not found! Please install psutil.")
        return

    if limconf.value is None:
        logger.warning("Timed limit %s is not set!", name)
        return

    if limconf.value <= 0:
        logger.warning("Timed limit %s is negative: %.1f", name, limconf.value)
        return

    logger.info("Starting %s timed check every %d second(s).", name,
                                                               limconf.period_s)
    lc = LoopingCall(deferToThread, _check)
    lc.start(limconf.period_s / 10)

    return lc



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

import re
import locale
import collections

from spyne import D
from random import choice
from contextlib import closing
from inspect import isclass, getargspec

from lxml import html
from lxml.builder import E

from spyne import Unicode, Decimal, Boolean, ComplexModelBase, Array, File, \
    ModelBase, AnyHtml, AnyUri, Integer
from spyne.util import six
from spyne.util.tdict import tdict
from spyne.util.oset import oset
from spyne.util.cdict import cdict
from spyne.util.six.moves.urllib.parse import urlencode

from spyne.protocol.html import HtmlBase
from spyne.protocol.html.table.column import HtmlColumnTableRowProtocol


def camel_case_to_uscore_gen(string):
    for i, s in enumerate(string):
        if s.isupper():
            if i > 0:
                yield "_"
            yield s.lower()
        else:
            yield s


camel_case_to_uscore = lambda s: ''.join(camel_case_to_uscore_gen(s))


def _gen_html5_epsilon():
    """See paragraph 17 at:

    https://www.w3.org/TR/2012/WD-html5-20121025/common-microsyntaxes.html
                                 #rules-for-parsing-floating-point-number-values
    """
    import math
    return D('1e%s' % int(math.log(2 ** (-1024)) / math.log(10)))


class HtmlFormWidget(HtmlBase):
    DEFAULT_INPUT_WRAPPER_CLASS = 'label-input-wrapper'
    DEFAULT_ANCHOR_WRAPPER_CLASS = 'label-anchor-wrapper'

    WRAP_FORWARD = type("WRAP_FORWARD", (object,), {})
    WRAP_REVERSED = type("WRAP_REVERSED", (object,), {})

    HTML5_EPSILON = _gen_html5_epsilon()
    HTML_FORM = 'form'
    HTML_INPUT = 'input'
    HTML_TEXTAREA = 'textarea'
    HTML_SELECT = 'select'
    HTML_OPTION = 'option'
    HTML_CHECKBOX_TAG = 'input'

    def __init__(self, app=None, encoding='utf8',
            ignore_uncap=False, ignore_wrappers=False,
                cloth=None, cloth_parser=None, polymorphic=True, hier_delim='.',
                     label=True, doctype=None, placeholder=None,
               input_class=None, input_div_class=None, input_wrapper_class=None,
                                                   label_class=None, type=None):

        super(HtmlFormWidget, self).__init__(app=app, encoding=encoding,
            doctype=doctype,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                cloth=cloth, cloth_parser=cloth_parser, polymorphic=polymorphic,
                                                          hier_delim=hier_delim)
        self.label = label
        self.placeholder = placeholder
        self._init_input_vars(input_class, input_div_class,
                                               input_wrapper_class, label_class)

        self.use_global_null_handler = False
        self.type = type

    def _init_input_vars(self, input_class, input_div_class,
                                              input_wrapper_class, label_class):
        self.input_class = input_class
        self.input_div_class = input_div_class
        self.input_wrapper_class = input_wrapper_class
        if self.input_wrapper_class is None:
            self.input_wrapper_class = self.DEFAULT_INPUT_WRAPPER_CLASS
        self.label_class = label_class

    def to_subprot(self, ctx, cls, inst, parent, name, subprot, **kwargs):
        if isinstance(subprot, HtmlFormWidget):
            if subprot.input_class is None:
                subprot.input_class = self.input_class

            if subprot.input_wrapper_class == self.DEFAULT_INPUT_WRAPPER_CLASS:
                subprot.input_wrapper_class = self.input_wrapper_class

            if subprot.label_class is None:
                subprot.label_class = self.label_class

        return super(HtmlFormWidget, self).to_subprot(ctx, cls, inst, parent,
                                                        name, subprot, **kwargs)

    @staticmethod
    def _format_js(lines, **kwargs):
        js = []
        for i, line in enumerate(lines):
            js.append(lines[i] % kwargs)

        return E.script("$(function(){\n\t%s\n});" % '\n\t'.join(js),
                                                        type="text/javascript")

    def _gen_input_hidden(self, cls, inst, parent, name, **kwargs):
        val = self.to_unicode(cls, inst)
        elt = E.input(type="hidden", name=name)
        if val is not None:
            elt.attrib['value'] = val
        parent.write(elt)

    def _gen_label_for(self, ctx, cls, name, input_id=None, **kwargs):
        attrib = {}

        if input_id is not None:
            attrib['for'] = input_id
        if self.label_class is not None:
            attrib['class'] = self.label_class

        return E.label(self.trc(cls, ctx.locale, name), **attrib)

    def _gen_label_wrapper_class(self, ctx, cls, name):
        classes = [self.input_wrapper_class, self.selsafe(name)]
        return {'class': ' '.join(classes)}

    def _wrap_with_label(self, ctx, cls, name, input_elt, no_label=False,
                                             wrap_label=WRAP_FORWARD, **kwargs):
        input_id = input_elt.attrib['id']
        if self.input_div_class is not None:
            input_elt = E.div(input_elt, **{'class': self.input_div_class})

        attrib = self._gen_label_wrapper_class(ctx, cls, name)

        cls_attrs = self.get_cls_attrs(cls)
        wants_no_label = cls_attrs.label is False or no_label or not self.label
        if wants_no_label:
            if wrap_label is None:
                retval = input_elt
            else:
                retval = E.div(input_elt, **attrib)

        else:
            label_attrib = {'for': input_id}
            if self.label_class is not None:
                label_attrib['class'] = self.label_class

            retval = E.label(self.trc(cls, ctx.locale, name), **label_attrib)
            if wrap_label is HtmlFormWidget.WRAP_FORWARD:
                retval = E.div(retval, input_elt, **attrib)

            elif wrap_label is HtmlFormWidget.WRAP_REVERSED:
                retval = E.div(input_elt, retval, **attrib)

            elif wrap_label is None:
                pass

            else:
                raise ValueError(wrap_label)

        return retval

    def _write_elt_with_label(self, ctx, cls, parent, name, elt):
        label = self._gen_label_for(ctx, cls, name)
        attrib = self._gen_label_wrapper_class(ctx, cls, name)
        with parent.element('div', attrib=attrib):
            parent.write(label)
            parent.write(elt)

    def _gen_input_elt_id(self, name, array_index=None, **kwargs):
        if array_index is None:
            return "%s_input" % (self.selsafe(name),)
        return "%s_%d_input" % (self.selsafe(name), array_index)

    def _gen_input_name(self, name, array_index=None):
        if array_index is None:
            return name
        return "%s[%d]" % (name, array_index)

    def _gen_input_attrs_novalue(self, ctx, cls, name, cls_attrs, **kwargs):
        elt_class = oset([
            camel_case_to_uscore(cls.get_type_name()),
            name.rsplit(self.hier_delim, 1)[-1],
            re.sub(r'\[[0-9]+\]', '', name).replace(self.hier_delim, '__'),
        ])

        if self.input_class is not None:
            elt_class.add(self.input_class)
        elt_class = ' '.join(elt_class)

        elt_attrs = tdict(six.string_types, six.string_types, {
            'id': self._gen_input_elt_id(name, **kwargs),
            'name': self._gen_input_name(name),
            'type': 'text',
            'class': elt_class,
        })

        if cls_attrs.pattern is not None:
            elt_attrs['pattern'] = cls_attrs.pattern

        if cls_attrs.read_only:
            elt_attrs['readonly'] = ""

        placeholder = cls_attrs.placeholder
        if placeholder is None:
            placeholder = self.placeholder

        if isinstance(placeholder, six.string_types):
            elt_attrs['placeholder'] = placeholder

        elif placeholder:
            elt_attrs['placeholder'] = self.trc(cls, ctx.locale, name)

        # Required bool means, in HTML context, a checkbox that needs to be
        # checked, which is not what we mean here at all.
        if not issubclass(cls, Boolean):
            # We used OR here because html forms send empty values anyway. So a
            # missing value is sent as null as well.
            if cls_attrs.min_occurs >= 1 or cls_attrs.nullable == False:
                elt_attrs['required'] = ''

        return elt_attrs

    def _gen_input_attrs(self, ctx, cls, inst, name, cls_attrs, **kwargs):
        elt_attrs = self._gen_input_attrs_novalue(ctx, cls, name, cls_attrs,
                                                                       **kwargs)

        if (cls_attrs.write_once and inst is not None) or \
                                                       cls_attrs.write is False:
            elt_attrs['readonly'] = ""

        val = self.to_unicode(cls, inst)
        if val is not None:
            elt_attrs['value'] = val

        return elt_attrs

    def _gen_input(self, ctx, cls, inst, name, cls_attrs, tag=None, **kwargs):
        elt_attrs = self._gen_input_attrs(ctx, cls, inst, name, cls_attrs,
                                                                       **kwargs)
        if tag is None:
            tag = self.HTML_INPUT

        values = cls_attrs.values

        if values is not None and len(values) > 0:
            tag = self.HTML_SELECT
            if 'value' in elt_attrs:
                del elt_attrs['value']
            if 'type' in elt_attrs:
                del elt_attrs['type']

        if cls_attrs.min_occurs == 1 and cls_attrs.nullable == False:
            elt = html.fromstring('<%s required>' % tag)
            elt.attrib.update(elt_attrs)
        else:
            elt = E(tag, **elt_attrs)

        if values is None or len(values) == 0:
            return elt

        return self._gen_options(ctx, cls, inst, name, cls_attrs, elt, **kwargs)

    def _gen_options(self, ctx, cls, inst, name, cls_attrs, elt, **kwargs):
        field_name = name.split('.')[-1]

        values_dict = cls_attrs.values_dict
        if values_dict is None:
            values_dict = {}

        inststr = self.to_unicode(cls, inst)
        if cls_attrs.write is False and inststr is not None:
            inst_label = values_dict.get(inst, inststr)
            if isinstance(inst_label, dict):
                inst_label = self.trd(inst_label, ctx.locale, field_name)

            logger.debug("\t\tinst %r label %r", inst_label, inst)
            self._append_option(elt, inst_label, inststr)

        else:
            ever_selected = False
            we_have_empty = False

            if cls_attrs.nullable or cls_attrs.min_occurs == 0:
                self._append_option(elt, '', value='')
                we_have_empty = True
                # if none are selected, this will be the default anyway, so
                # no need to add selected attribute to the option tag.

            # FIXME: cache this!
            i = 0
            for i, v in enumerate(cls_attrs.values):
                selected = False
                valstr = self.to_unicode(cls, v)
                if valstr is None:
                    valstr = ""

                if inst == v:
                    ever_selected = True
                    selected = True

                val_label = values_dict.get(v, valstr)
                logger.debug("\t\tother values inst %r label %r", v, val_label)
                if isinstance(val_label, dict):
                    val_label = self.trd(val_label, ctx.locale, valstr)

                self._append_option(elt, val_label, valstr,
                                                     selected=selected, index=i)

            if not (ever_selected or we_have_empty):
                self._append_option(elt, label='', value='', selected=True,
                                                                        index=i)

            return elt

    def _append_option(self, parent, label, value, selected=False, index=-1):
        attrib = dict(value=value)
        if selected:
            attrib['selected'] = ''

        parent.append(E(self.HTML_OPTION, label, **attrib))

    def _gen_input_textarea(self, ctx, cls, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_attrs = self._gen_input_attrs_novalue(ctx, cls, name, cls_attrs)
        if cls_attrs.min_occurs == 1 and cls_attrs.nullable == False:
            elt = html.fromstring('<textarea required>')
            elt.attrib.update(elt_attrs)
        else:
            elt = E.textarea(**elt_attrs)

        return cls_attrs, elt

    def _gen_input_unicode(self, ctx, cls, inst, name, attr_override={},
                                                                      **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        max_len = attr_override.get('max_len', cls_attrs.max_len)
        min_len = attr_override.get('min_len', cls_attrs.min_len)
        values = attr_override.get('values', cls_attrs.values)

        if len(values) == 0 and max_len >= D('inf'):
            tag = self.HTML_TEXTAREA

            elt_attrs = self._gen_input_attrs_novalue(ctx, cls, name, cls_attrs)
            if (cls_attrs.write_once and inst is not None) or \
                                                       cls_attrs.write is False:
                elt_attrs['readonly'] = ""

            if cls_attrs.min_occurs == 1 and cls_attrs.nullable == False:
                elt = html.fromstring('<%s required>' % tag)
                elt.attrib.update(elt_attrs)
            else:
                elt = E(tag, **elt_attrs)

            if not (inst is None and isinstance(inst, type)):
                elt.text = self.to_unicode(cls, inst)

        else:
            elt = self._gen_input(ctx, cls, inst, name, cls_attrs, **kwargs)
            elt.attrib['type'] = 'text'

            if max_len < Unicode.Attributes.max_len:
                elt.attrib['maxlength'] = str(int(max_len))

            if min_len > Unicode.Attributes.min_len:
                elt.attrib['minlength'] = str(int(min_len))

        return cls_attrs, elt

    @staticmethod
    def _apply_number_constraints(cls_attrs, elt, epsilon):
        if cls_attrs.max_str_len != Decimal.Attributes.max_str_len:
            elt.attrib['maxlength'] = str(cls_attrs.max_str_len)

        if cls_attrs.ge != Decimal.Attributes.ge:
            elt.attrib['min'] = str(cls_attrs.ge)

        if cls_attrs.gt != Decimal.Attributes.gt:
            elt.attrib['min'] = str(cls_attrs.gt + epsilon)

        if cls_attrs.le != Decimal.Attributes.le:
            elt.attrib['max'] = str(cls_attrs.le)

        if cls_attrs.lt != Decimal.Attributes.lt:
            elt.attrib['max'] = str(cls_attrs.lt - epsilon)

        if elt.attrib.get('min', None) is None \
                                            and cls_attrs.min_bound is not None:
            elt.attrib['min'] = str(cls_attrs.min_bound)

        if elt.attrib.get('max', None) is None \
                                            and cls_attrs.max_bound is not None:
            elt.attrib['max'] = str(cls_attrs.max_bound)

    def _switch_to_prot_type(self, cls, inst):
        if self.type is not None and not (cls is self.type):
            cls = self.type
            if len(self.type_attrs) > 0:
                cls = self.type.customize(**self.type_attrs)

        return cls, inst

    def complex_model_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        newcls, newinst = self._switch_to_prot_type(cls, inst)
        if newcls is cls:
            return self.not_supported(ctx, cls)
        return self.to_parent(ctx, newcls, newinst, parent, name, **kwargs)


class ConditionalRendererBase(HtmlFormWidget):
    def __init__(self):
        super(ConditionalRendererBase, self).__init__()

        self.serialization_handlers = cdict({
            ModelBase: self.model_base_to_parent,
        })

    def switch_to_subprot(self, ctx, cls, inst, parent, name, subprot=None,
                                                                      **kwargs):
        if subprot is None:
            subprot = ctx.protocol.prot_stack[-2]

        logger.debug("Subprot switch from %r back to parent prot %r",
                                                                  self, subprot)

        return self.to_subprot(ctx, cls, inst, parent, name, subprot, **kwargs)

    def model_base_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        if self.cond(ctx, cls, inst):
            return self.true(ctx, cls, inst, parent, name, **kwargs)
        else:
            return self.false(ctx, cls, inst, parent, name, **kwargs)

    def cond(self, ctx, cls, inst):
        return choice([True, False])

    def true(self, ctx, cls, inst, parent, name, **kwargs):
        return self.switch_to_subprot(ctx, cls, inst, parent, name, **kwargs)

    def false(self, ctx, cls, inst, parent, name, **kwargs):
        return self.switch_to_subprot(ctx, cls, inst, parent, name, **kwargs)


# TODO: Make label optional
class PasswordWidget(HtmlFormWidget):
    def __init__(self, *args, **kwargs):
        super(PasswordWidget, self).__init__(*args, **kwargs)

        self.serialization_handlers = cdict({
            Unicode: self.unicode_to_parent,
            ComplexModelBase: self.complex_model_to_parent,
        })

    def unicode_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs, elt = self._gen_input_unicode(ctx, cls, inst, name, **kwargs)
        elt.attrib['type'] = 'password'
        parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))


class HrefWidget(HtmlFormWidget):
    supported_types = (Unicode, Decimal)

    def __init__(self, href=None, hidden_input=False, label=True, quote=None,
                                                             anchor_class=None):
        """Render current object (inst) as an anchor (the <a> tag)

        :param href: Base HREF for Anchor. Can only be None when
            inst is an instance of AnyUri.Value
        :param hidden_input: If True, generate hidden <input> with inst as value
        :param label: If True, Generate <label> element.
        :param quote: If not None, should be a callable like ``quote_plus``
            to make sure arbitrary strings get properly escaped as query string
            parameters.
        :param anchor_class: If not None, goes into the "class" attribute of
            the <a> tag.
        """

        super(HrefWidget, self).__init__(label=label)

        self.href = href
        self.hidden_input = hidden_input
        self.anchor_class = anchor_class
        self.quote = quote

        self.serialization_handlers = cdict({
            ModelBase: self.model_base_to_parent,
            AnyUri: self.any_uri_to_parent,
            ComplexModelBase: self.not_supported,
        })

    def model_base_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        assert self.href is not None

        anchor_str = self.to_unicode(cls, inst)
        if anchor_str is None:
            anchor_str = ''

        try:
            if self.quote is not None:
                inst = self.quote(inst)

            href = self.href.format(inst)

        except Exception as e:
            logger.warning("Error generating href: %r", e)
            href = self.href

        self.render_anchor(ctx, cls, inst, parent, name, anchor_str, href,
                                                                      **kwargs)

    def render_anchor(self, ctx, cls, inst, parent, name, anchor_str, href,
                                                                      **kwargs):
        elt = E.a(anchor_str)
        if href is not None:
            elt.attrib['href'] = href

        if self.anchor_class is not None:
            elt.attrib['class'] = self.anchor_class

        if self.label:
            label = self._gen_label_for(ctx, cls, name)
            attrib = self._gen_label_wrapper_class(ctx, cls, name)

            with parent.element('div', attrib=attrib):
                parent.write(label)

                span_attrib = {}
                parent.write(E.span(elt, **span_attrib))

        else:
            parent.write(elt)

        cls_attr = self.get_cls_attrs(cls)
        if self.hidden_input and (inst is not None or cls_attr.min_occurs >= 1):
            self._gen_input_hidden(cls, inst, parent, name, **kwargs)

    def any_uri_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        retval = self.gen_anchor(cls, inst, name, self.anchor_class)
        parent.write(retval)


class ParentHrefWidget(HrefWidget):
    """Render current object as a link using information from its parent
    object.

    An example href value:
        "/some_slug?id={0.id}&name={0.name}

    **FIXME:** This needs to be made to use urlencode instead of plain string
    formatting.
    """

    supported_types = (Unicode, Decimal)

    def model_base_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        anchor_str = self.to_unicode(cls, inst)
        if anchor_str is None:
            anchor_str = ''

        inst_stack = ctx.protocol.inst_stack
        if len(inst_stack) < 2:
            logger.warning("No parent instance found.")
            href = self.href

        else:
            _, parent_inst, _ = inst_stack[-2]

            try:
                href = self.href.format(parent_inst)
            except Exception as e:
                logger.warning("Error generating href: %r", e)
                href = self.href

        self.render_anchor(ctx, cls, inst, parent, name, anchor_str, href,
                                                                      **kwargs)


class SimpleRenderWidget(HtmlFormWidget):
    def __init__(self, label=True, type=None, hidden=False, elt=None):
        super(SimpleRenderWidget, self).__init__(label=label)

        self.elt = elt
        self.type = type
        self.hidden = hidden
        self.serialization_handlers = cdict({
            ModelBase: self.model_base_to_parent,
            AnyHtml: self.any_html_to_parent,
            AnyUri: self.any_uri_to_parent,
            ComplexModelBase: self.complex_model_to_parent,
        })

    def _gen_text_str(self, cls, inst, **kwargs):
        text_str = self.to_unicode(cls, inst, **kwargs)

        if text_str is None:
            text_str = ''

            cls_attr = self.get_cls_attrs(cls)
            if cls_attr.min_occurs == 0:
                return None

        return text_str

    def _wrap_with_label_simple(self, ctx, cls, text_str, parent, name):
        label = self._gen_label_for(ctx, cls, name)
        attrib = self._gen_label_wrapper_class(ctx, cls, name)

        with parent.element('div', attrib=attrib):
            parent.write(label)

            span_attrib = {}
            parent.write(E.span(text_str, **span_attrib))

    def model_base_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        newcls, newinst = self._switch_to_prot_type(cls, inst)
        if not (newcls is cls):
            return self.to_parent(ctx, newcls, newinst, parent, name, **kwargs)

        text_str = self._gen_text_str(cls, inst, **kwargs)
        if text_str is None:
            return

        if self.label:
            self._wrap_with_label_simple(ctx, cls, text_str, parent, name)
        else:
            if self.elt is not None:
                parent.write(E(self.elt.tag, text_str, **self.elt.attrib))
            else:
                parent.write(text_str)

        if self.hidden:
            self._gen_input_hidden(cls, inst, parent, name, **kwargs)

    def any_uri_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        newcls, newinst = self._switch_to_prot_type(cls, inst)
        if not (newcls is cls):
            return self.to_parent(ctx, newcls, newinst, parent, name, **kwargs)

        if isinstance(inst, AnyUri.Value):
            href_str = self._gen_text_str(cls, inst.href)
            link_str = inst.text
        else:
            href_str = self._gen_text_str(cls, inst)
            if href_str is None:
                return
            link_str = href_str

        anchor = E.a(link_str, href=href_str)

        if self.label:
            self._wrap_with_label_simple(ctx, cls, anchor, parent, name)
        else:
            parent.write(anchor)

        if self.hidden:
            self._gen_input_hidden(cls, inst, parent, name, **kwargs)

    def any_html_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        newcls, newinst = self._switch_to_prot_type(cls, inst)
        if not (newcls is cls):
            return self.to_parent(ctx, newcls, newinst, parent, name, **kwargs)

        if isinstance(inst, six.string_types):
            inst = html.fromstring(inst)

        if inst is None:
            inst = ''

            cls_attr = self.get_cls_attrs(cls)
            if cls_attr.min_occurs == 0:
                return

        if self.label:
            label = self._gen_label_for(ctx, cls, name)
            attrib = self._gen_label_wrapper_class(ctx, cls, name)

            with parent.element('div', attrib=attrib):
                parent.write(label)
                parent.write(inst)

        else:
            parent.write(inst)

        if self.hidden:
            self._gen_input_hidden(cls, inst, parent, name, **kwargs)


class ComplexRenderWidget(HtmlFormWidget):
    type_attrs = dict(validate_freq=False)

    def __init__(self, text_field=None, id_field=None, type=None,
           hidden_fields=None, label=True, null_str='[NULL]', input_class=None):
        """A widget base that renders complex objects as simple html elements.

        :param text_field: The name of the field containing a human readable
            string that represents the object.
        :param id_field: The name of the field containing the unique identifier
            of the object.
        :param type: If not `None`, overrides the object type being rendered.
            Useful for e.g. combining multiple fields to one field.
        :param hidden_fields: A sequence of field names that will be rendered as
            hidden <input> tags.
        :param label: If ``True``, a ``<label>`` tag is generated for the
            relevant widget id.
        """

        super(ComplexRenderWidget, self).__init__(label=label,
                                                        input_class=input_class)

        if not six.PY2:
            str_types = (str,)
        else:
            str_types = (str, unicode)

        self.text_field = text_field
        if isinstance(self.text_field, str_types):
            self.text_field = tuple([s for s in self.text_field.split('.')
                                                                 if len(s) > 0])

        self.id_fields = id_field
        if isinstance(self.id_fields, str_types):
            self.id_fields = tuple([s for s in self.id_fields.split('.')
                                                                 if len(s) > 0])

        self.type = type
        self.hidden_fields = hidden_fields
        self.null_str = null_str

        self.serialization_handlers = cdict({
            ComplexModelBase: self.complex_model_to_parent,
        })

    def _get_type(self, cls, field_name):
        assert len(field_name) > 0

        for field_name_fragment in field_name:
            fti = cls.get_flat_type_info(cls)
            cls = fti[field_name_fragment]

        return cls

    def _get_value_str(self, inst_type, inst, field_name):
        assert len(field_name) > 0

        for field_name_fragment in field_name:
            inst = getattr(inst, field_name_fragment)

        return self.to_unicode(inst_type, inst)

    def _prep_inst(self, cls, inst, fti):
        id_name = id_type = id_str = None
        if self.id_fields is not None:
            id_name = self.id_fields
            id_type = self._get_type(cls, id_name)

        text_str = text_type = None
        text_name = self.text_field
        if text_name is not None:
            text_str = self.null_str
            text_type = self._get_type(cls, text_name)

        if inst is not None:
            if id_name is not None:
                id_str = self._get_value_str(id_type, inst, id_name)

            if text_name is not None:
                text_str = self._get_value_str(text_type, inst, text_name)

        if id_str is None:
            id_str = ""
        if text_str is None:
            text_str = ""

        return id_str, text_str

    def _write_hidden_fields(self, ctx, cls, inst, parent, name, fti, **kwargs):
        if self.hidden_fields is not None:
            for key in self.hidden_fields:
                self._gen_input_hidden(fti[key], getattr(inst, key, None),
                            parent, self.hier_delim.join((name, key)), **kwargs)

    def complex_model_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls, inst = self._switch_to_prot_type(cls, inst)
        fti = cls.get_flat_type_info(cls)
        _, text_str = self._prep_inst(cls, inst, fti)

        if self.label:
            elt_id = self._gen_input_elt_id(name, **kwargs)
            label = self._gen_label_for(ctx, cls, name, elt_id)
            attrib = self._gen_label_wrapper_class(ctx, cls, name)
            with parent.element('div', attrib=attrib):
                parent.write(label)
                parent.write(text_str)

        else:
            parent.write(text_str)

        self._write_hidden_fields(ctx, cls, inst, parent, name, fti, **kwargs)


class ComplexHrefWidget(ComplexRenderWidget, HtmlColumnTableRowProtocol):
    def __init__(self, text_field, id_field, type=None, hidden_fields=None,
                   empty_widget=None, label=True, url=None, id_field_name=None):
        """Widget that renders complex objects as links. Hidden fields are
        skipped then the given instance has the value of `id_field` is `None`.

        Please see :class:`ComplexRenderWidget` docstring for more info.

        :param empty_widget: The widget to be used when the value of the
            instance to be rendered is `None`.
        """
        super(ComplexHrefWidget, self).__init__(text_field, id_field,
                            type=type, hidden_fields=hidden_fields, label=label)

        self.id_field_name = id_field_name
        self.empty_widget = empty_widget
        self.url = url

        if isclass(empty_widget):
            assert issubclass(empty_widget, ComplexRenderWidget), "I don't know" \
                         "how to instantiate a non-ComplexRenderWidget-subclass"

            self.empty_widget = empty_widget(self.text_field, self.id_fields,
                      others=True, others_order_by=self.text_field, label=False)

    def column_table_before_row(self, ctx, cls, inst, parent, name, **kwargs):
        # the ctxstack here is lxml element context, has nothing to do with
        # spyne contexts.

        ctxstack = getattr(ctx.protocol[self], 'array_subprot_ctxstack', [])

        tr_ctx = parent.element('tr')
        tr_ctx.__enter__()
        ctxstack.append(tr_ctx)

        td_ctx = parent.element('td')
        td_ctx.__enter__()
        ctxstack.append(td_ctx)

        ctx.protocol[self].array_subprot_ctxstack = ctxstack

    def column_table_after_row(self, ctx, cls, inst, parent, name, **kwargs):
        ctxstack = ctx.protocol[self].array_subprot_ctxstack

        for elt_ctx in reversed(ctxstack):
            elt_ctx.__exit__(None, None, None)

        del ctxstack[:]

    def complex_model_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        fti = cls.get_flat_type_info(cls)
        id_str, text_str = self._prep_inst(cls, inst, fti)

        attrib = {}
        label_div_attrib = {}

        label = None
        if self.label:
            label = self._gen_label_for(ctx, cls, name, "")
            label_div_attrib = {
                'class': '%s %s' % (
                        self.DEFAULT_ANCHOR_WRAPPER_CLASS, cls.get_type_name())
            }

        if id_str != "":
            tn_url = self.url
            if tn_url is None:
                tn = cls.get_type_name()
                tn_url = camel_case_to_uscore(tn)

            id_field_name = self.id_field_name
            if id_field_name is None:
                id_field_name = '.'.join(self.id_fields)

            attrib['href'] = tn_url + "?" + urlencode({id_field_name: id_str})

            a_tag = E.a(text_str, **attrib)
            if self.label:
                parent.write(E.div(label, a_tag, **label_div_attrib))
            else:
                parent.write(a_tag)

            self._write_hidden_fields(ctx, cls, inst, parent, name, fti, **kwargs)

        elif self.empty_widget is not None:
            if self.label:
                with parent.element('div', attrib=label_div_attrib):
                    parent.write(label)
                    self.empty_widget \
                              .to_parent(ctx, cls, inst, parent, name, **kwargs)
            else:
                self.empty_widget \
                              .to_parent(ctx, cls, inst, parent, name, **kwargs)


# FIXME: We need a better explanation for the very simple thing that
# override_parent does.
class SelectWidgetBase(ComplexRenderWidget):
    def __init__(self, text_field, id_field, hidden_fields=None, label=True,
                   type=None, inst_type=None, others=None, others_order_by=None,
                     override_parent=False, nonempty_widget=ComplexRenderWidget,
                                                              input_class=None):
        """Widget that renders complex objects as <select> tags.

        Please see :class:`ComplexRenderWidget` docstring for more info.

        :param others: When `True` fetches all values from the corresponding
            persistence backend entity and adds them as options.

        :param others_order_by: When not `None`, requests an ordered result set
            from the database order by the field name(s) given in this field.
            If given as a string, it's treated as one argument whereas given
            as a list or a tuple of strings, it's treated as multiple field
            names.

        :param override_parent: Overrides parent's name with the names from
            ComboBox.

        :param nonempty_widget: The widget to be used for non-null instances of
            classes with ``write_once`` attribute set.

        :param type: Force class type to given type.

        :param inst_type: Also force instance type to given type. Defaults to
            whatever passed in as ``type``.

        :param html_class: When not none, pass as class attribute to the
            ``<input>`` or ``<select>`` element.
        """

        super(SelectWidgetBase, self).__init__(id_field=id_field,
                             text_field=text_field, hidden_fields=hidden_fields,
                                label=label, type=type, input_class=input_class)

        self.others = others
        if callable(others):
            argspec = getargspec(others)
            assert argspec.varargs is not None or len(argspec.args) >= 2, \
                "The callable passed to 'others' must accept at least 2 " \
                "arguments: Instances of 'MethodContext' and 'ComboBoxWidget'."

        self.others_order_by = others_order_by
        self.override_parent = override_parent
        self.inst_type = inst_type
        if inst_type is None:
            self.inst_type = type

        self.nonempty_widget = nonempty_widget
        if isclass(nonempty_widget):
            assert issubclass(nonempty_widget, ComplexRenderWidget), \
                         "I don't know" \
                         "how to instantiate a non-ComplexRenderWidget-subclass"

            self.nonempty_widget = nonempty_widget(self.text_field,
                    id_field=self.id_fields, label=self.label, type=self.type,
                         hidden_fields=(self.hidden_fields or ()) + (id_field,))

        self.serialization_handlers[ModelBase] = self.model_base_to_parent

    def _write_empty(self, parent):
        parent.write(E(self.HTML_OPTION))

    def _write_select(self, ctx, cls, inst, parent, name, fti, **kwargs):
        raise NotImplementedError()

    def _write_select_impl(self, ctx, cls, tag_attrib, data, fti, parent):
        attr = self.get_cls_attrs(cls)

        with parent.element(self.HTML_SELECT, attrib=tag_attrib):
            if self.others is None:
                for v_id_str, v_text_str in data:
                    elt = E(self.HTML_OPTION, v_text_str, value=v_id_str,
                                                                    selected="")
                    parent.write(elt)
                return

            we_have_empty = False
            if attr.min_occurs == 0:
                self._write_empty(parent)
                we_have_empty = True

            is_callable = callable(self.others)
            is_iterable = isinstance(self.others, collections.Iterable)
            is_autogen = (self.others is True)

            if not any((is_callable, is_iterable, is_autogen)):
                return

            selected = False
            if is_autogen:
                logger.debug("Auto-generating combobox contents for %r", cls)
                db = ctx.app.config.stores['sql_main'].itself

                # FIXME: this blocks the reactor
                with closing(db.Session()) as session:
                    # Get class
                    subval_cls = cls.__orig__
                    if subval_cls is None:
                        subval_cls = cls

                    q = session.query(cls.__orig__ or cls)

                    # Apply ordering parameters if possible
                    oob = self.others_order_by
                    oob_id = id(oob)
                    if oob is not None:
                        if not isinstance(oob, (list, tuple)):
                            oob = (oob,)

                        q = q.order_by(*oob)

                    cache_key = (cls.__orig__ or cls, oob_id)
                    selected = self._write_options(ctx, cls, parent, fti, q,
                                                                data, cache_key)

            elif is_iterable or is_callable:
                insts = self.others
                if is_callable:
                    logger.debug("Generating select contents from callable "
                                                                  "for %r", cls)
                    insts = self.others(ctx, self)

                else:
                    logger.debug("Generating select contents from iterable "
                                                                  "for %r", cls)

                selected = self._write_options(ctx, cls, parent, fti, insts,
                                                           data, cache_key=None)

            else:
                raise Exception("This should not be possible")

            if not (we_have_empty or selected):
                self._write_empty(parent)

    def _write_options(self, ctx, cls, parent, fti, insts, data, cache_key):
        selected = False
        data = set((i for (i,t) in data))

        # Generate cache entry
        cache_entry = None
        if cache_key is not None:
            objcache = ctx.protocol.objcache
            cache_entry = objcache.get(cache_key, None)
            if cache_entry is not None:
                logger.debug("Using cache key %r to fill <select> for %r",
                                                                 cache_key, cls)
                for id_str, text_str in cache_entry:
                    elt = E(self.HTML_OPTION, text_str, value=id_str)

                    if id_str in data:
                        elt.attrib['selected'] = ""
                        selected = True

                    parent.write(elt)

                return selected

            logger.debug("Generated cache entry for key %r to fill <select> "
                                                       "for %r", cache_key, cls)

            cache_entry = objcache[cache_key] = []

        # FIXME: this iteration blocks the reactor
        for o in insts:
            id_str, text_str = self._prep_inst(cls, o, fti)

            elt = E(self.HTML_OPTION, text_str, value=id_str)
            if id_str in data:
                elt.attrib['selected'] = ""
                selected = True

            parent.write(elt)
            if cache_entry is not None:
                cache_entry.append( (id_str, text_str) )

        return selected

    def model_base_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attr = self.get_cls_attrs(cls)

        parent_cls, parent_inst, parent_fromarr = ctx.protocol.inst_stack[-2]

        if self.override_parent and cls is not parent_cls:
            logger.debug("ComboBoxWidget.override_parent "
                                        "cls switch: %r => %r", cls, parent_cls)

            parent_name = name.rsplit(self.hier_delim, 1)[0]
            kwargs['from_arr'] = parent_fromarr

            if inst is not None and self.nonempty_widget is not None:
                return self.nonempty_widget.to_parent(ctx,
                         parent_cls, parent_inst, parent, parent_name, **kwargs)

            return self.complex_model_to_parent(ctx,
                         parent_cls, parent_inst, parent, parent_name, **kwargs)

        if cls_attr.write_once and inst is not None:
            return self._gen_input_hidden(cls, inst, parent, name, **kwargs)

        return self.not_supported(ctx, cls, inst, parent, name, **kwargs)

    def complex_model_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attr = self.get_cls_attrs(cls)

        if inst is not None and cls_attr.write_once:
            return self.nonempty_widget.to_parent(ctx, cls, inst, parent,
                                                                 name, **kwargs)

        if self.type is not None:
            logger.debug("ComboBoxWidget.type cls switch: %r => %r",
                                                                 cls, self.type)
            cls = self.type

            if len(self.type_attrs) > 0:
                cls = self.type.customize(**self.type_attrs)

        if self.inst_type is not None:
            logger.debug("ComboBoxWidget.type inst switch: %r => %r",
                                                             cls, self.type)

            inst = self.type.init_from(inst)

        self.cm_to_parent_impl(ctx, cls, inst, parent, name, **kwargs)

    def cm_to_parent_impl(self, ctx, cls, inst, parent, name, **kwargs):
        fti = cls.get_flat_type_info(cls)

        if self.label:
            elt_id = self._gen_input_elt_id(name, **kwargs)
            label = self._gen_label_for(ctx, cls, name, elt_id)
            attrib = self._gen_label_wrapper_class(ctx, cls, name)
            with parent.element('div', attrib=attrib):
                parent.write(label)
                self._write_select(ctx, cls, inst, parent, name, fti, **kwargs)

        else:
            self._write_select(ctx, cls, inst, parent, name, fti, **kwargs)

        self._write_hidden_fields(ctx, cls, inst, parent, name, fti, **kwargs)


class ComboBoxWidget(SelectWidgetBase):
    def _write_select(self, ctx, cls, inst, parent, name, fti, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        v_id_str, v_text_str = self._prep_inst(cls, inst, fti)

        sub_name = self.hier_delim.join((name,) + self.id_fields)
        elt_attrs = self._gen_input_attrs_novalue(ctx, cls, sub_name, cls_attrs,
                                                                       **kwargs)
        data = ((v_id_str, v_text_str),)

        self._write_select_impl(ctx, cls, elt_attrs, data, fti, parent)


class MultiSelectWidget(SelectWidgetBase):
    supported_types = (Array,)

    def cm_to_parent_impl(self, ctx, cls, inst, parent, name, **kwargs):
        if issubclass(cls, Array):
            cls = next(iter(cls._type_info.values()))

        super(MultiSelectWidget, self).cm_to_parent_impl(ctx, cls, inst,
                                                         parent, name, **kwargs)

    def _write_select(self, ctx, cls, inst, parent, name, fti, **kwargs):
        cls_attr = self.get_cls_attrs(cls)

        data = []
        for subinst in inst:
            data.add(self._prep_inst(cls, subinst, fti))

        if self.override_parent:
            name = name.rsplit(self.hier_delim)[0]

        sub_name = self.hier_delim.join((name,) + self.id_fields)
        tag_attrib = self._gen_input_attrs_novalue(ctx, cls, sub_name, cls_attr,
                                                                       **kwargs)

        tag_attrib['multiple'] = ""
        self._write_select_impl(ctx, cls, tag_attrib, data, fti, parent)

    def _write_empty(self, parent):
        pass


class SimpleReadableNumberWidget(SimpleRenderWidget):
    def __init__(self, label=True, type=None, hidden=False,
                                                           locale='en_US.utf8'):
        super(SimpleReadableNumberWidget, self).__init__(
                                          label=label, type=type, hidden=hidden)

        self.locale = locale

        self.serialization_handlers = cdict({
            Decimal: self.decimal_to_parent,
            Integer: self.integer_to_parent,
            ComplexModelBase: self.complex_model_to_parent,
        })

    def write_number(self, ctx, cls, inst, parent, name, fstr, **kwargs):
        if inst is None:
            return

        locale.setlocale(locale.LC_ALL, self.locale)
        valstr = locale.format(fstr, inst, grouping=True)

        if self.label:
            label = self._gen_label_for(ctx, cls, name)
            attrib = self._gen_label_wrapper_class(ctx, cls, name)
            with parent.element('div', attrib=attrib):
                parent.write(label)
                parent.write(valstr)

        else:
            parent.write(valstr)

        if self.hidden:
            self._gen_input_hidden(cls, inst, parent, name, **kwargs)

    def integer_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls, inst = self._switch_to_prot_type(cls, inst)
        cls_attrs = self.get_cls_attrs(cls)

        fstring = cls_attrs.format
        if fstring is None:
            fstring = "%d"

        self.write_number(ctx, cls, inst, parent, name, fstring, **kwargs)

    def decimal_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls, inst = self._switch_to_prot_type(cls, inst)
        cls_attrs = self.get_cls_attrs(cls)

        fstring = cls_attrs.format
        if fstring is None:
            fstring = "%f"

            if cls_attrs.fraction_digits == D('inf'):
                fd = 2  # FIXME: chosen by fair dice roll
            else:
                fd = int(cls_attrs.fraction_digits)

            if fd is not None:
                fstring = "%%.%df" % fd

        self.write_number(ctx, cls, inst, parent, name, fstring, **kwargs)


class JQFileUploadWidget(SimpleRenderWidget):
    def __init__(self, label=True):
        super(JQFileUploadWidget, self).__init__(label=label)

        self.serialization_handlers = cdict({
            File: self.file_to_parent,
        })

    def file_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        pass


class TrueFalseWidget(SimpleRenderWidget):
    SYM_TRUE = u"✔"
    SYM_FALSE = u"✘"
    SYM_NONE = u"●"

    def __init__(self, label=True, type=None, hidden=False, center=False,
                        color=True, color_true='green', color_false='red',
                        color_none='gray', display='inline-block', width="100%",
                                            addtl_css="text-decoration: none;"):
        super(TrueFalseWidget, self).__init__(label=label, type=type,
                                                                  hidden=hidden)

        self.center = center
        self.color = color

        self.color_true = color_true
        self.color_false = color_false
        self.color_none = color_none

        self.display = display
        self.width = width
        self.addtl_css = addtl_css

    def get_none(self):
        if self.color:
            return E.span(self.SYM_NONE,
                style="{}color:{}".format(self.addtl_css, self.color_none),
                **{"class": "widget-none widget-none-color"}
            )

        else:
            return E.span(self.SYM_NONE,
                style=self.addtl_css,
                **{"class": "widget-none widget-none-dull"}
            )

    def get_true(self):
        if self.color:
            return E.span(self.SYM_TRUE,
                style="{}color:{}".format(self.addtl_css, self.color_true),
                **{"class": "widget-true widget-true-color"}
            )

        else:
            return E.span(self.SYM_TRUE,
                style=self.addtl_css,
                **{"class": "widget-true widget-true-dull"}
            )

    def get_false(self):
        if self.color:
            return E.span(self.SYM_FALSE,
                style="{}color:{}".format(self.addtl_css, self.color_false),
                **{"class": "widget-false widget-false-color"}
            )

        else:
            elt = E.span(self.SYM_FALSE,
                style=self.addtl_css,
                **{"class": "widget-false widget-false-dull"}
            )

    def to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        if self.type is not None:
            cls = self.type

        if inst is None:
            elt = self.get_none()

        elif inst:
            elt = self.get_true()

        else:
            elt = self.get_false()

        styles = ["background: transparent"]
        if self.width:
            styles.append("width: 100%")
        if self.display:
            styles.append("display: %s" % (self.display,))
        if self.center:
            styles.append("text-align:center;")

        elt = E.div(elt, style=';'.join(styles))

        if self.label:
            label = self._gen_label_for(ctx, cls, name)
            attrib = self._gen_label_wrapper_class(ctx, cls, name)
            with parent.element('div', attrib=attrib):
                parent.write(label)
                parent.write(elt)

        else:
            parent.write(elt)

        if self.hidden:
            self._gen_input_hidden(cls, inst, parent, name, **kwargs)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


FILE_VERSION_KEY = 'file_version'
STATIC_DESC_ROOT = "Directory that contains static files for the root url."
STATIC_DESC_URL = "Directory that contains static files for the url '%s'."


class ServiceDisabled(Exception):
    pass


from neurons.daemon.config._base import LOGLEVEL_MAP
from neurons.daemon.config._base import LOGLEVEL_MAP_ABB
from neurons.daemon.config._base import LOGLEVEL_STR_MAP
from neurons.daemon.config._base import ServiceDefinition

from neurons.daemon.config.endpoint import Service
from neurons.daemon.config.endpoint import Client
from neurons.daemon.config.endpoint import Server
from neurons.daemon.config.endpoint import SslServer
from neurons.daemon.config.endpoint import HttpServer
from neurons.daemon.config.endpoint import WsgiServer
from neurons.daemon.config.endpoint import HttpApplication
from neurons.daemon.config.endpoint import StaticFileServer

from neurons.daemon.config.store import FileStore
from neurons.daemon.config.store import LdapStore
from neurons.daemon.config.store import RelationalStore

from neurons.daemon.config.daemon import Daemon
from neurons.daemon.config.daemon import ServiceDaemon
from neurons.daemon.config.daemon import EmailAlert
from neurons.daemon.config.daemon import AlertDestination



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

# Exit codes
EXIT_SUCCESS = 0
EXIT_ERR_UNKNOWN = 1
# EXIT_SIGNAL = 100
EXIT_ERR_MEMORY = 201
EXIT_ERR_LISTEN_TCP = 100000
EXIT_ERR_LISTEN_UDP = 200000


from neurons.daemon import dowser
from neurons.daemon.main import main

from neurons.daemon.config import Service
from neurons.daemon.config import Client
from neurons.daemon.config import Server
from neurons.daemon.config import SslServer
from neurons.daemon.config import HttpServer
from neurons.daemon.config import WsgiServer
from neurons.daemon.config import ServiceDaemon
from neurons.daemon.config import HttpApplication
from neurons.daemon.config import StaticFileServer

from neurons.daemon.config import FileStore
from neurons.daemon.config import LdapStore
from neurons.daemon.config import RelationalStore

from neurons.daemon.config import Daemon
from neurons.daemon.config import ServiceDaemon
from neurons.daemon.config import EmailAlert
from neurons.daemon.config import AlertDestination

from neurons.daemon.config import ServiceDefinition


config_data = None
"""The last parsed ``Daemon`` instance."""



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

from spyne import M, Integer32, Integer64, Integer, ByteArray, Unicode, \
    DateTime, Boolean, IpAddress, AnyXml, AnyDict, ComplexModel, Uuid

from neurons import TableModel
from neurons.version import Version

from sqlalchemy import sql


class LogEntryMixin(ComplexModel):
    __mixin__ = True
    __table_args__ = {"sqlite_autoincrement": True}

    _type_info = [
        ('id', Integer64(primary_key=True)),
        ('id_session', Uuid),
        ('id_message', Uuid),

        ('time', M(DateTime(timezone=False, server_default=sql.func.now()))),
        ('domain', Unicode(255)),
        ('username', Unicode(255)),

        ('method_name', M(Unicode(255))),
        ('duration_ms', Integer32),
        ('read_only', M(Boolean)),

        ('daemon_version', Integer64),
        ('daemon_name', Unicode(32)),

        ('is_request', M(Boolean(default=True, server_default='true'))),
        ('err_code_in', Unicode),
        ('err_code_out', Unicode),

        ('host', IpAddress),

        # TODO: figure out how these will work
        # ('id_session', Uuid),
        # ('id_message', Uuid),

        ('data_in_xml', AnyXml),
        ('data_in_json', AnyDict(store_as='json')),
        ('data_out_int', Integer64),
        ('data_out_xml', AnyXml),
        ('data_out_json', Integer64),
    ]


def migrate_2(config, session):
    session.connection().execute("""
      alter table neurons_log add column id_session uuid;
      alter table neurons_log add column id_message uuid;
    """)


migdict = {
    2: migrate_2,
}


DB_SCHEMA_VERSION = max(migdict.keys())


def TLogEntry(table_model=TableModel):
    # Register version table for migration
    Version.register_submodule("neurons_log", migdict, DB_SCHEMA_VERSION)

    class LogEntry(table_model, LogEntryMixin):
        __namespace__ = 'http://spyne.io/neurons/log'
        __tablename__ = 'neurons_log'

    return LogEntry



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

import os
import sys
import gzip
import shutil
import neurons

from time import time
from pprint import pformat
from datetime import date

from spyne import ComplexModel
from spyne import Unicode
from spyne.util import six
from spyne.util.color import R, B, YEL, DARK_R, DARK_G


LOGLEVEL_MAP = dict(zip(
    ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
    [logging.DEBUG, logging.INFO, logging.WARNING, logging.ERROR,
                                                               logging.CRITICAL]
))

LOGLEVEL_STR_MAP = {v: k for k, v in LOGLEVEL_MAP.items()}

LOGLEVEL_MAP_ABB = {
    logging.DEBUG: 'D',
    logging.INFO: B('I'),
    logging.WARNING: YEL('W'),
    logging.ERROR: R('E'),
    logging.CRITICAL: DARK_R('C'),
}


def _get_reactor_thread_sigil(record):
    if neurons.REACTOR_THREAD_ID is None:
        return ' '

    if record.thread == neurons.REACTOR_THREAD_ID:
        return DARK_R('R')

    return DARK_G('P')


class Logger(ComplexModel):
    path = Unicode
    level = Unicode(values=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])

    def __init__(self, *args, **kwargs):
        self._parent = None

        super(Logger, self).__init__(*args, **kwargs)

    def set_parent(self, parent):
        assert parent is not None
        self._parent = parent

    def apply(self):
        if self.path in (None, '', '.'):
            _logger = logging.getLogger()
        else:
            _logger = logging.getLogger(self.path)

        _logger.setLevel(LOGLEVEL_MAP[self.level])

        if self.path == '.':
            self._parent.boot_message()
            logger.info("Root logger level override: %s", self.level)
        else:
            logger.info("Logger level override %s = %s", self.path, self.level)

        return self


def TTwistedHandler(config, loggers, _meminfo):
    from twisted.logger import LogLevel

    # this is supposed to override the Logger object above. that's not cool but
    # that's the way it has to be
    from twisted.logger import Logger

    LOGLEVEL_TWISTED_MAP = {
        logging.DEBUG: LogLevel.debug,
        logging.INFO: LogLevel.info,
        logging.WARN: LogLevel.warn,
        logging.ERROR: LogLevel.error,
        logging.CRITICAL: LogLevel.critical,
    }

    class TwistedHandler(logging.Handler):
        if config.log_rss:
            if _meminfo is None:
                @staticmethod
                def _modify_record(record):
                    record.msg = '[psutil?] %s' % record.msg
            else:
                @staticmethod
                def _modify_record(record):
                    meminfo = _meminfo()
                    record.msg = '[%.2f] %s' % \
                                         (meminfo.rss / 1024.0 ** 2, record.msg)

        else:
            def _modify_record(self, record):
                pass

        def emit(self, record):
            assert isinstance(record, logging.LogRecord)

            record.l = LOGLEVEL_MAP_ABB.get(record.levelno, "?")
            record.r = _get_reactor_thread_sigil(record)

            self._modify_record(record)

            _logger = loggers.get(record.name, None)
            if _logger is None:
                _logger = loggers[record.name] = Logger(record.name)

            if six.PY2 and hasattr(record, 'msg') \
                                                and isinstance(record.msg, str):
                record.msg = record.msg.decode('utf8')

            t = self.format(record)

            if six.PY2 and isinstance(t, str):
                t = t.decode('utf8', errors='replace')

            _logger.emit(LOGLEVEL_TWISTED_MAP[record.levelno], log_text=t)

    return TwistedHandler


def TDynamicallyRotatedLog(config, comp_method):
    from twisted.python.logfile import DailyLogFile

    class DynamicallyRotatedLog(DailyLogFile):
        def suffix(self, tupledate):
            # this just adds leading zeroes to dates. it's otherwise
            # same with parent
            try:
                return '-'.join(("%02d" % i for i in tupledate))
            except:
                # try taking a float unixtime
                return '-'.join(("%02d" % i for i in
                    self.toDate(tupledate)))

        def write(self, data):
            if isinstance(data, six.text_type):
                data = data.encode('utf8')

            DailyLogFile.write(self, data)

        if comp_method == 'gzip':
            def compress_rotated_file(self, file_name):
                start = time()
                target_file_name = '{}.gz'.format(file_name)

                with open(file_name, 'rb') as f_in, \
                    gzip.open(target_file_name, 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)

                os.unlink(file_name)

                logger.info(
                    "Rotated log file '%s' => '%s', took %f seconds.",
                    file_name, target_file_name, time() - start)
        else:
            def compress_rotated_file(self, file_name):
                pass

        def rotate(self):
            """Rotate the file and create a new one.

            If it's not possible to open new logfile, this will fail
            silently, and continue logging to old logfile.
            """

            # copied from base class and modified with call to
            # compress_rotated_file

            if not (os.access(self.directory, os.W_OK) and os.access(
                self.path, os.W_OK)):
                return

            newpath = "%s.%s" % (self.path, self.suffix(self.lastDate))

            if os.path.exists(newpath):
                newpath_tmpl = "%s_{}" % newpath
                i = 0
                while os.path.exists(newpath):
                    i += 1

                    newpath = newpath_tmpl.format(i)

            self._file.close()

            os.rename(self.path, newpath)
            self._openFile()

            if 'twisted' in sys.modules:
                from twisted.internet.threads import deferToThread

                deferToThread(self.compress_rotated_file, newpath) \
                    .addErrback(lambda err:
                                         logger.error("%s", err.getTraceback()))

            else:
                self.compress_rotated_file(newpath)

        if config.logger_dest_rotation_period == "DAILY":
            def shouldRotate(self):
                return self.toDate() != self.lastDate

        elif config.logger_dest_rotation_period == "WEEKLY":
            def shouldRotate(self):
                today = date(*self.toDate())
                last = date(*self.lastDate)
                return (today.year != last.year or
                        today.isocalendar()[1] != last.isocalendar()[1])

        elif config.logger_dest_rotation_period == "MONTHLY":
            def shouldRotate(self):
                return self.toDate()[:2] != self.lastDate[:2]

        else:
            def shouldRotate(self):
                logger.warning("Invalid logger_dest_rotation_period value %r",
                                             config.logger_dest_rotation_period)
                return False

    return DynamicallyRotatedLog


def Trecord_as_string(formatter):
    from twisted.logger import LogLevel

    LOGLEVEL_TWISTED_MAP = {
        logging.DEBUG: LogLevel.debug,
        logging.INFO: LogLevel.info,
        logging.WARN: LogLevel.warn,
        logging.ERROR: LogLevel.error,
        logging.CRITICAL: LogLevel.critical,
    }

    TWISTED_LOGLEVEL_MAP = {v: k for k, v in LOGLEVEL_TWISTED_MAP.items()}

    def record_as_string(record):
        if 'log_failure' in record:
            failure = record['log_failure']
            try:
                s = pformat(vars(failure.value))
            except TypeError:
                # vars() argument must have __dict__ attribute
                s = repr(failure.value)
            return "%s: %s" % (failure.type, s)

        if 'log_text' in record:
            return record['log_text'] + "\n"

        if 'log_format' in record:
            level = record.get('log_level', LogLevel.debug)
            level = LOGLEVEL_MAP_ABB[TWISTED_LOGLEVEL_MAP[level]]

            text = record['log_format'].format(**record) + "\n"
            ns = record.get('log_namespace', "???")
            lineno = 0
            record = logging.LogRecord('?', level, ns, lineno, text,
                None, None)
            record.l = level
            record.r = _get_reactor_thread_sigil(record)
            record.module = ns.split('.')[-2]

            return formatter.format(record)

        if 'log_io' in record:
            return record['log_io'] + "\n"

        if 'message' in record:
            return record['message'] + "\n"

        return pformat(record)

    return record_as_string



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from neurons.form.widget import HtmlFormWidget
from neurons.form.widget import HrefWidget
from neurons.form.widget import ParentHrefWidget
from neurons.form.widget import PasswordWidget
from neurons.form.widget import TrueFalseWidget

from neurons.form.widget import ComboBoxWidget
from neurons.form.widget import MultiSelectWidget

from neurons.form.widget import SimpleRenderWidget
from neurons.form.widget import ConditionalRendererBase
from neurons.form.widget import ComplexRenderWidget
from neurons.form.widget import SimpleReadableNumberWidget

from neurons.form.widget import ComplexHrefWidget

from neurons.form.form import Tab
from neurons.form.form import Fieldset
from neurons.form.form import HtmlFormRoot
from neurons.form.form import THtmlFormRoot
from neurons.form.form import HtmlForm

from neurons.form.table import HtmlFormTable

def tou(name):
    if name is not None:
        return name


#!/usr/bin/env python
# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import logging
logging.basicConfig(level=logging.DEBUG)

import re
import unittest

from neurons.form import HtmlFormTable, HtmlForm
from neurons.form.test import strip_ns
from neurons.form.const import T_TEST

from spyne.util.test import show
from spyne.util.color import R
from spyne import Application, NullServer, Service, rpc, ComplexModel, \
    Integer, Array, Unicode
from lxml import etree, html


def _test_type(cls, inst, prot_cls=HtmlFormTable):
    from spyne.util import appreg; appreg.applications.clear()

    class SomeService(Service):
        @rpc(_returns=cls, _body_style='bare')
        def some_call(ctx):
            return inst

    prot = prot_cls(cloth=T_TEST, doctype='<!DOCTYPE html>', label=False)
    app = Application([SomeService], 'some_ns', out_protocol=prot)

    null = NullServer(app, ostr=True)

    ret = ''.join(null.service.some_call())
    try:
        elt = html.fromstring(ret)
    except:
        print(ret)
        raise

    show(elt)
    elt = elt.xpath('//form')[0]
    elt = strip_ns(elt)  # get rid of namespaces to simplify xpaths in tests

    print(R("========== fragment =========="))
    print(etree.tostring(elt, pretty_print=True))
    print(R("========== fragment =========="))

    return elt


class TestFormTable(unittest.TestCase):
    def test_simple_array(self):
        v = range(5)
        elt = _test_type(Array(Integer), v)

        assert elt.xpath('table/tbody/tr/td/div/input/@value') == \
                                                       ['0', '1', '2', '3', '4']
        assert elt.xpath('table/tbody/tr/td/button/text()') == ['-'] * 5 + ['+']
        for i, name in enumerate(elt.xpath('div/div/input/@name')):
            assert re.match(r'ints\[0*%d\]' % i, name)

    def test_complex_array(self):
        v = "domates"  # 🍅

        class SomeObject(ComplexModel):
            i = Integer
            s = Unicode(64)

        v = [SomeObject(i=i, s=s) for i, s in enumerate(v)]
        elt = _test_type(Array(SomeObject), v)

        assert elt.xpath(
            'table/tbody/tr/td/div/input[contains(@class, "i")]/@value') == \
                                                           [str(o.i) for o in v]
        assert elt.xpath(
            'table/tbody/tr/td/div/input[contains(@class, "s")]/@value') == \
                                                           [str(o.s) for o in v]

        assert elt.xpath('table/tbody/tr/td/button/text()') == ['-'] * len(v) + ['+']
        for i, name in enumerate(elt.xpath('div/div/input/@name')):
            assert re.match(r'ints\[0*%d\]' % i, name)

    def test_simple_sub(self):
        v = "domates"  # 🍅

        class SomeObject(ComplexModel):
            i = Integer
            s = Array(Unicode(64), prot=HtmlFormTable(label=False), label=True)

        v = SomeObject(i=42, s=v)
        elt = _test_type(SomeObject, v, HtmlForm)

        assert elt.xpath(
            'fieldset/div/table/tbody/tr/td/div'
                            '/input[contains(@class, "s")]/@value') == \
                                                                [s for s in v.s]
        assert elt.xpath(
            'fieldset/div/input[contains(@class, "i")]/@value') == [str(v.i)]

        assert elt.xpath(
            'fieldset/div/table/tbody/tr/td/button/text()') == ['-'] * len(v.s) + ['+']

    def test_unicode_null(self):
        v = None
        elt = _test_type(Unicode(64), v).xpath('div/input')[0]

        assert elt.attrib['type'] == 'text'
        assert not ('value' in elt.attrib)


if __name__ == '__main__':
    unittest.main()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from twisted.web.static import File
from twisted.web.resource import ForbiddenResource
from twisted.python.filepath import InsecurePath

from spyne.server.twisted.http import get_twisted_child_with_default


def TCheckedFile(disallowed_exts, url):
    class CheckedFile(File):
        def __init__(self, *args, **kwargs):
            File.__init__(self, *args, **kwargs)

            self.prepath = url

        def getChildWithDefault(self, path, request):
            return get_twisted_child_with_default(self, path, request)

        def child(self, path):
            retval = File.child(self, path)

            if path.rsplit(".", 1)[-1] in disallowed_exts:
                raise InsecurePath("%r is disallowed." % (path,))

            return retval

    return CheckedFile


def TStaticFile(disallowed_exts, url):
    class StaticFile(TCheckedFile(disallowed_exts, url)):
        def directoryListing(self):
            return ForbiddenResource()

    return StaticFile



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from __future__ import print_function, absolute_import

import logging
logger = logging.getLogger(__name__)

import os

from os.path import abspath

from spyne import ComplexModel, Boolean, Unicode, \
    UnsignedInteger16, M, Decimal, UnsignedInteger
from spyne.util import get_version

from neurons.daemon.cli import config_overrides
from neurons.daemon.store import SqlDataStore, LdapDataStore


class StorageInfo(ComplexModel):
    name = Unicode
    backend = Unicode

    def __init__(self, *args, **kwargs):
        self._parent = None

        super(StorageInfo, self).__init__(*args, **kwargs)

    def set_parent(self, parent):
        assert self._parent is None
        assert parent is not None

        self._parent = parent

    def _parse_overrides(self):
        pass

    def close(self):
        pass


class PoolConfig(ComplexModel):
    type = Unicode


class AsyncPoolConfig(ComplexModel):
    type = Unicode(values=['txpostgres'])


class SyncPoolConfig(ComplexModel):
    type = Unicode(values=[
            'SingletonThreadPool', 'QueuePool', 'NullPool',
            'StaticPool', 'AssertionPool'
        ])


DistinguishedName = Unicode


class LdapStore(StorageInfo):
    method = M(Unicode(values=['simple', 'gssapi']))
    backend = M(Unicode(values=['python-ldap']))

    host = M(Unicode)
    port = UnsignedInteger16(default=389)
    base_dn = DistinguishedName
    bind_dn = DistinguishedName
    password = Unicode
    timeout = Decimal(gt=0, default=10)
    version = UnsignedInteger(default=3, vaues=[2, 3])
    use_tls = Boolean(default=False)
    referrals = Boolean(default=False)

    def __init__(self, *args, **kwargs):
        super(LdapStore, self).__init__(*args, **kwargs)
        self.itself = None

    def apply(self):
        if self.backend in LdapDataStore.SUPPORTED_BACKENDS:
            self.itself = LdapDataStore(self)
            self.itself.apply()

        else:
            raise ValueError(self.backend)


class FileStore(StorageInfo):
    path = M(Unicode)

    def apply(self):
        self.path = abspath(self.path)
        if not os.path.isdir(self.path):
            os.makedirs(self.path)


class RelationalStore(StorageInfo):
    # this is not supposed to be mandatory because it's overrideable by cli args
    conn_str = Unicode

    # move these to QueuePool config.
    pool_size = UnsignedInteger(default=10)
    pool_recycle = UnsignedInteger(default=3600)
    pool_timeout = UnsignedInteger(default=30)
    pool_use_lifo = Boolean(default=False)
    pool_pre_ping = Boolean(default=True)

    max_overflow = UnsignedInteger(default=3)
    echo_pool = Boolean(default=False)

    sync_pool = Boolean(default=True)
    sync_pool_type = Unicode(
        default='QueuePool',
        values=[
            'SingletonThreadPool', 'QueuePool', 'NullPool',
            'StaticPool', 'AssertionPool'
        ],
    )

    async_pool = Boolean(default=True)

    def _parse_overrides(self):
        super(RelationalStore, self)._parse_overrides()

        config_keys = self.config_keys()
        ovs = set(config_overrides).intersection(set(config_keys))
        for c in ovs:
            setattr(self, config_keys[c], config_overrides[c])
            del config_overrides[c]


    def config_keys(self):
        return {'--store-{}-{}'.format(self.name, k.replace("_", "-")): k
                               for k in self.get_flat_type_info(self.__class__)}

    def __init__(self, *args, **kwargs):
        super(RelationalStore, self).__init__(*args, **kwargs)
        self.itself = None

    def apply(self):
        if self.name is None:
            raise ValueError("Unnamed store entry")

        if self.conn_str is None:
            raise ValueError("Connection string is None for store %s"
                                                                  % (self.name))

        kwargs = dict(
            pool_size=self.pool_size,
            echo_pool=self.echo_pool,
            pool_timeout=self.pool_timeout,
            pool_recycle=self.pool_recycle,
            max_overflow=self.max_overflow,
            logging_name=self.name,
        )

        if get_version('sqlalchemy')[0:2] >= (1, 2):
            kwargs['pool_pre_ping'] = self.pool_pre_ping

        if get_version('sqlalchemy')[0:2] >= (1, 3):
            kwargs['pool_use_lifo'] = self.pool_use_lifo

        self.itself = SqlDataStore(self.name, self.conn_str, **kwargs)

        if not (self.async_pool or self.sync_pool):
            logger.debug("Store '%s' is disabled.", self.name)

        if self.async_pool:
            if self.conn_str.startswith('postgres'):
                self.itself.add_txpool()
            else:
                self.async_pool = False

        if not self.sync_pool:
            self.itself.Session = None
            self.itself.metadata = None
            self.itself.engine.dispose()
            self.itself.engine = None

        return self

    def close(self):
        if self.async_pool:
            self.itself.txpool.close()

        if self.sync_pool:
            self.itself.Session = None
            self.itself.metadata = None
            self.itself.engine.dispose()
            self.itself.engine = None

        self.itself = None

        return self



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

from neurons import Application
from neurons.daemon.config import HttpServer, StaticFileServer
from neurons.daemon.dowser.const import ASSETS_DIR
from neurons.daemon.dowser.service import DowserServices

from spyne import Double
from spyne.protocol.html import HtmlBase
from spyne.protocol.http import HttpRpc


class DowserServer(HttpServer):
    tick_period_sec = Double(default=60)

    @property
    def host(self):
        from neurons.daemon.ipc import get_own_dowser_address
        host, port = get_own_dowser_address()
        return host

    @host.setter
    def host(self, _):
        pass

    @property
    def port(self):
        from neurons.daemon.ipc import get_own_dowser_address
        host, port = get_own_dowser_address()
        return port

    @port.setter
    def port(self, _):
        pass


def start_dowser(config):
    from twisted.internet.task import LoopingCall
    from twisted.internet.threads import deferToThread

    subconfig = config.services['dowser']
    if subconfig.subapps is None:
        subconfig.subapps = {}

    subconfig.subapps[''] = \
        Application(
            [
                DowserServices,
            ],
            tns='neurons.daemon', name='Dowser',
            in_protocol=HttpRpc(validator='soft'),
            out_protocol=HtmlBase(),
            config=config,
        )

    subconfig.subapps['assets'] = \
        StaticFileServer(
            url='assets',
            path=ASSETS_DIR,
            list_contents=False,
            disallowed_exts=["py"]
        )

    task = LoopingCall(deferToThread, DowserServices.tick)
    task.start(subconfig.tick_period_sec)

    return subconfig.gen_site()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

import argparse
import os.path

from gettext import gettext
from decimal import Decimal as D

from spyne import Boolean, Unicode, Integer, Decimal, ComplexModelBase, Array
from spyne.protocol import ProtocolBase
from spyne.util import six
from spyne.util.cdict import cdict


ARGTYPE_MAP = cdict({
    Integer: int,
    Decimal: D,
})


config_overrides = dict()
"""Wildcard command line arguments that don't fit into argparse's fixed
argument keys model."""


class ArgumentParser(argparse.ArgumentParser):
    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)

        if len(argv) > 0:
            newargv = []
            for a in argv:
                if a.startswith("--assets-"):
                    opt, val = a.split("=", 2)
                    val = os.path.abspath(val)

                elif a.startswith("--host-") or a.startswith("--port-"):
                    opt, val = a.split("=", 2)

                elif a.startswith("--store-"):
                    opt, val = a.split("=", 2)

                else:
                    newargv.append(a)
                    continue

                config_overrides[opt] = val

            argv = newargv

        if len(argv) > 0:
            msg = gettext('unrecognized arguments: %s')
            self.error(msg % ' '.join(argv))

        print("Overrides:", config_overrides)

        return args


def Tmust_exist(test_func):
    """Generic for an argparse type.

    :param test_func: One of :func:`os.path.isdir`, :func:`os.path.isfile` or
            :func:`os.path.ismount`.
    """

    def must_exist(file_name):
        if file_name:
            file_name = os.path.abspath(file_name)
            if not test_func(file_name):
                raise OSError("%r is invalid." % file_name)
            return file_name

    return must_exist


def enum(*values):
    """Enum type for argparse.

    :param values: Values that the option at hand can have. Must all be unicode.
    """

    def enum_(value):
        if value:
            if not (value in values):
                raise ValueError('Argument must be one of %r' % values)
            return value

    return enum_


file_must_exist = Tmust_exist(os.path.isfile)
dir_must_exist = Tmust_exist(os.path.isdir)


def _is_array_of_primitives(cls):
    if issubclass(cls, Array):
        subcls, = cls._type_info.values()
        return not issubclass(subcls, ComplexModelBase)

    elif cls.Attributes.max_occurs > 1:
        return not issubclass(cls, ComplexModelBase)

    return False


def _is_array_of_complexes(cls):
    if issubclass(cls, Array):
        subcls, = cls._type_info.values()
        return issubclass(subcls, ComplexModelBase)

    elif cls.Attributes.max_occurs > 1:
        return issubclass(cls, ComplexModelBase)

    return False


_some_prot = ProtocolBase()

def spyne_to_argparse(cls, ignore_defaults):
    fti = cls.get_flat_type_info(cls)
    parser = ArgumentParser(description=cls.__doc__, add_help=False)

    parser.add_argument('-c', '--config-file', type=os.path.abspath,
                                             help="An alternative config file.")

    for k, v in sorted(fti.items(), key=lambda i: i[0]):
        attrs = _some_prot.get_cls_attrs(v)
        if attrs.no_cli:
            continue

        args = ['--%s' % k.replace('_', '-')]
        if attrs.short is not None:
            args.append('-%s' % attrs.short)

        assert not ('-c' in args or '--config-file' in args)

        kwargs = {}

        if attrs.help is not None:
            kwargs['help'] = attrs.help

        if attrs.metavar is not None:
            kwargs['metavar'] = attrs.metavar

        if ignore_defaults:
            kwargs['default'] = argparse.SUPPRESS

        elif attrs.default is not None:
            kwargs['default'] = attrs.default

        # types
        if _is_array_of_primitives(v):
            kwargs['nargs'] = "+"

        elif _is_array_of_complexes(v):
            continue

        elif issubclass(v, Boolean):
            kwargs['action'] = "store_const"
            kwargs['const'] = True

        elif issubclass(v, Unicode):
            if len(v.Attributes.values) > 0:
                kwargs['type'] = enum(*v.Attributes.values)
            else:
                kwargs['type'] = six.text_type

        elif issubclass(v, tuple(ARGTYPE_MAP.keys())):
            kwargs['type'] = ARGTYPE_MAP[v]

        parser.add_argument(*args, **kwargs)
        logger.debug("add_argument %r %r", args, kwargs)

    return parser



#!/bin/env python
# encoding: utf-8

import os
import re

from setuptools import setup
from setuptools import find_packages


ROOT_NAME = 'garage'

v = open(os.path.join(os.path.dirname(__file__), ROOT_NAME, '__init__.py'))
VERSION = re.compile(r".*__version__ = '(.*?)'", re.S).match(v.read()).group(1)


setup(
    name=ROOT_NAME,
    version=VERSION,
    packages=find_packages(),

    install_requires=['neurons>=0.5'],
    entry_points={
        'console_scripts': [
            'garage_daemon=garage.main:main_daemon',
        ],
    },

    package_data={

    },

    author='Burak Arslan',
    author_email='burak.arslan@arskom.com.tr',

    description="Example Garage Management Project",
)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

from collections import defaultdict


class ReadContext(object):
    def __init__(self, parent):
        self.trusted = False

        self.parent = parent

        self.domain = None
        self.username = None

        self.log_entry = None

        self.do_log = True
        self.do_log_inbound = True
        self.do_log_outbound = False
        self.is_read_only = True

        self.short_log_level = logging.INFO

        self.sqla_sessions = defaultdict(list)

        self._initialized = True

    def __repr__(self):
        if self.domain is None and self.username is None:
           return "Context(ro={})".format(self.is_read_only)

        if self.domain is None:
           return "Context(ro={}, user={})".format(self.is_read_only,
                                                                  self.username)

        return "Context(ro={}, user={}@{})".format(self.is_read_only,
                                                     self.username, self.domain)

    def sqla_finalize(self, session):
         pass

    def get_du(self):
        if self.domain is None:
            return self.username
        return "%s@%s" % (self.username, self.domain)

    def get_main_session(self, **kwargs):
        return self.get_session(self.parent.app.config.get_main_store(**kwargs))

    def get_session(self, store, **kwargs):
        if store.type == 'sqlalchemy':
            sessions = self.sqla_sessions[id(store)]

            if len(sessions) == 0:
                session = store.Session(**kwargs)
                self.sqla_sessions[id(store)].append(session)

            else:
                assert len(kwargs) == 0
                session = sessions[0]

            return session

        elif store.type == 'ldap':
            return store.conn

        raise NotImplementedError(store)

    def close(self, no_error=True):
        for sessions in self.sqla_sessions.values():
            for session in sessions:
                if no_error:
                    self.sqla_finalize(session)
                session.close()

        # TODO: Close LDAP sessions?

    def has_role(self, role_cls):
        for role in self.roles:
            if role.role == role_cls.__role_name__:
                return True

        return False


class WriteContext(ReadContext):
    def __init__(self, parent):
        super(WriteContext, self).__init__(parent)

        self.is_read_only = False

    def sqla_finalize(self, session):
        logger.debug("Committing transaction for ctx 0x%012X", id(self.parent))
        session.commit()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


from spyne.util.odict import odict

from neurons.daemon.config import ServiceDisabled


class wdict(odict):
    def getwrite(self, key, *args):
        if len(args) > 0:
            if not key in self:
                self[key], = args
        return self[key]


class wrdict(wdict):
    def getwrite(self, key, *args):
        """Raises ServiceDisabled when the service has disabled == True"""

        retval = super(wrdict, self).getwrite(key, *args)

        if getattr(retval, 'disabled', None):
            raise ServiceDisabled(getattr(retval, 'name', "??"))

        return retval


def Twrdict(parent, keyattr=None):
    class twrdict(wrdict):
        if keyattr is not None:
            def __setitem__(self, key, value):
                super(wrdict, self).__setitem__(key, value)
                setattr(value, keyattr, key)
                value._parent = parent
        else:
            def __setitem__(self, key, value):
                super(wrdict, self).__setitem__(key, value)
                value._parent = parent

    return twrdict



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

"""
Implementation of the GeoJson in terms of Spyne types.

See http://geojson.org/ for more info.

You need to use ``JsonDocument(polymorphic=True)`` as your out_protocol to make
sure the output is proper json.

Here's also how to use it in a non-RPC context:

>>> from spyne.util.dictdoc import get_object_as_json
>>> from neurons.geojson import Feature, Point
>>> f = Feature(
...     geometry=Point(Position(1,2)),
...     properties=dict(key="value"),
... )
>>> print(get_object_as_json(f, Feature, complex_as=dict, polymorphic=True))
{
    "type": "Feature",
    "geometry": {
        "type": "Point",
        "coordinates": [1, 2]
    },
    "properties": {"key": "value"}
}
"""


from spyne import Unicode, ComplexModel, Double, Array, AnyDict


class Geometry(ComplexModel):
    def __init__(self, coordinates):
        ComplexModel.__init__(self)

        self.coordinates = coordinates


class Position(ComplexModel):
    class Attributes(ComplexModel.Attributes):
        serialize_as = list

    _type_info = [
        ('x', Double),
        ('y', Double),
    ]

    def __init__(self, x, y):
        ComplexModel.__init__(self)

        self.x = x
        self.y = y


class Point(Geometry):
    _type_info = [
        ('type', Unicode(default="Point")),
        ('coordinates', Position),
    ]


class MultiPoint(Geometry):
    _type_info = [
        ('type', Unicode(default="MultiPoint")),
        ('coordinates', Array(Position)),
    ]


class LineString(Geometry):
    _type_info = [
        ('type', Unicode(default="LineString")),
        ('coordinates', Array(Position)),
    ]


class MultiLineString(Geometry):
    _type_info = [
        ('type', Unicode(default="MultiLineString")),
        ('coordinates', Array(Array(Position))),
    ]


class Polygon(Geometry):
    _type_info = [
        ('type', Unicode(default="Polygon")),
        ('coordinates', Array(Array(Position))),
    ]


class MultiPolygon(Geometry):
    _type_info = [
        ('type', Unicode(default="MultiPolygon")),
        ('coordinates', Array(Array(Array(Position)))),
    ]


class GeometryCollection(ComplexModel):
    _type_info = [
        ('type', Unicode(default="GeometryCollection")),
        ('geometries', Array(Geometry)),
    ]


class Feature(ComplexModel):
    _type_info = [
        ('type', Unicode(default="Feature")),
        ('geometry', Geometry),
        ('properties', AnyDict),
    ]


class FeatureCollection(ComplexModel):
    _type_info = [
        ('type', Unicode(default="FeatureCollection")),
        ('features', Array(Feature)),
    ]



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

import json

from lxml.html.builder import E

from slimit import mangler
from slimit.visitors.minvisitor import ECMAMinifier

from spyne import ComplexModel, Array, Unicode, XmlAttribute, AnyUri, \
    XmlData, Integer, UnsignedInteger, ComplexModelBase
from spyne.protocol.html import HtmlCloth
from spyne.store.relational import get_pk_columns
from spyne.util.dictdoc import get_object_as_simple_dict
from spyne.util.six.moves.urllib.parse import urlencode


SETUP_DATATABLES = """
neurons.setup_datatables = function(selector, data, hide) {
    var $table = $(selector);
    if (hide) {
        neurons.hide_empty_columns($table);
    }
    if ($table.length) {
        setTimeout(function () {
            $table.dataTable(data);
        }, 1);
    }
};
"""

HIDE_EMPTY_COLUMNS = """
neurons.hide_empty_columns = function ($table) {
    $($table.find("tr")[1]).find("td").each(function (_, elt) {
        setTimeout(function () {
            var cc = $(elt).attr("class").split(" ")[0];
            if ($table.find("td." + cc).text() == "") {
                $table.find("td." + cc).hide();
                $table.find("th." + cc).hide();
            }
        }, 0);
    });
};
"""

# http://stackoverflow.com/a/17772086
TYPE_CHECKERS = """
['Arguments', 'Function', 'String', 'Number', 'Date', 'RegExp'].forEach(
    function(name) {
        neurons['is' + name] = function(obj) {
              return toString.call(obj) == '[object ' + name + ']';
    };
});
"""

# http://stackoverflow.com/questions/728360#728694
CLONE = """
neurons.clone = function(obj) {
    var copy;

    // Handle the 3 simple types, and null or undefined
    if (null == obj || "object" != typeof obj) return obj;

    // Handle Date
    if (obj instanceof Date) {
        copy = new Date();
        copy.setTime(obj.getTime());
        return copy;
    }

    // Handle Array
    if (obj instanceof Array) {
        copy = [];
        for (var i = 0, len = obj.length; i < len; i++) {
            copy[i] = neurons.clone(obj[i]);
        }
        return copy;
    }

    // Handle Object
    if (obj instanceof Object) {
        copy = {};
        for (var attr in obj) {
            if (obj.hasOwnProperty(attr)) copy[attr] = neurons.clone(obj[attr]);
        }
        return copy;
    }

    throw new Error("Unable to copy obj! Its type isn't supported.");
};
"""

URLENCODE = """
neurons.urlencode_str = function(str) {
    var output = '';
    var x = 0;
    str = str.toString();
    var regex = /(^[a-zA-Z0-9_.]*)/;

    while (x < str.length) {
        var match = regex.exec(str.substr(x));

        if (match != null && match.length > 1 && match[1] != '') {
            output += match[1];
            x += match[1].length;
        }
        else {
            if (str[x] == ' ') {
                output += '+';
            }
            else {
                var charCode = str.charCodeAt(x);
                var hexVal = charCode.toString(16);
                output += '%' + ( hexVal.length < 2 ? '0' : '' ) + hexVal.toUpperCase();
            }
            x++;
        }
    }

    return output;
};

neurons.urlencode = function(obj) {
    var i = 0;
    var retval = "";
    for (var k in obj) if (obj.hasOwnProperty(k) && obj[k]) {
        if (i > 0) {
            retval += "&";
        }
        retval += k + '=' + neurons.urlencode_str(obj[k]);

        ++i;
    }
    return retval;
};

"""

class Link(ComplexModel):
    href = XmlAttribute(AnyUri)
    rel = XmlAttribute(Unicode(values=["stylesheet", 'import']))


class CascadingStyleSheet(ComplexModel):
    id = Unicode
    data = XmlData(Unicode)
    type = XmlAttribute(Unicode(values=["text/css"]))
    media = XmlAttribute(Unicode(values=["text/css"]))
    title = XmlAttribute(Unicode)


class ScriptElement(ComplexModel):
    id = XmlAttribute(Unicode)
    data = XmlData(Unicode)
    type = XmlAttribute(Unicode(values=["text/javascript"]))
    href = XmlAttribute(Unicode)


class ViewRenderer(HtmlCloth):
    def to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        if inst is None:
            return

        in_message = ctx.descriptor.in_message
        if not issubclass(in_message, ComplexModelBase):
            return

        view_names = tuple((k for k, v in in_message._type_info.items()
                                                    if issubclass(v, ViewBase)))
        if len(view_names) == 0:
            return

        view_name, = view_names
        path = ctx.transport.get_path()
        qs_dict = dict(ctx.in_body_doc.items())
        if name == 'prev':
            for k, v in qs_dict.items():
                if k.endswith('.start'):
                    del qs_dict[k]
                    break

        elif name == 'next':
            for k, v in qs_dict.items():
                if k.endswith('.end'):
                    del qs_dict[k]
                    break

        qs_dict.update(get_object_as_simple_dict(inst, prefix=(view_name,)))

        anchor_text = name
        anchor_href = "{}?{}".format(path, urlencode(qs_dict))

        parent.write(E.a(anchor_text, href=anchor_href))


class ViewBase(ComplexModel):
    LIMIT_MAX = 100
    START_MIN = 0

    _type_info = [
        ('end', Integer),
        ('start', Integer),
        ('limit', UnsignedInteger),
        ('offset', UnsignedInteger),
    ]

    def __init__(self, *args, **kwargs):
        self._limit = None

        super(ViewBase, self).__init__(*args, **kwargs)

    @property
    def limit(self):
        if self._limit is None:
            return None

        return min(self._limit, self.LIMIT_MAX)

    @limit.setter
    def limit(self, limit):
        self._limit = limit

    def apply(self, ctx, cls, q):
        if self is None:
            return q

        # FIXME: Support compound columns
        # FIXME: Actually we don't need the primary key. Just a unique column
        #        should be enough
        (pk_field_name, pk_field_type), = get_pk_columns(cls)
        pk_field = getattr(cls, pk_field_name)

        start = self.start
        if start is not None:
            q = q.filter(pk_field > start)

        end = self.end
        if end is not None:
            q = q.filter(pk_field < end)

        if self.sort_params is not None and len(self.sort_params) > 0:
            order_by = []

            for param in self.sort_params:
                field = getattr(cls, param.column)
                if param.is_descending():
                    field = field.desc()

                order_by.append(field)
                logger.debug("Order by %s", param.column)

            q = q.order_by(*order_by)

        else:
            logger.debug("Order by pk")
            if end is not None:
                q = q.order_by(pk_field.desc())
            else:
                q = q.order_by(pk_field)

        limit = self.limit
        if limit is not None:
            if limit > self.LIMIT_MAX:
                limit = self.LIMIT_MAX

            logger.debug("Limit %d", limit)
            q = q.limit(limit)

        offset = self.offset
        if offset is not None:
            logger.debug("Offset %d", offset)
            q = q.offset(offset)

        return q


class ScreenBase(ComplexModel):
    JQUERY_URL = "https://code.jquery.com/jquery-1.12.4.min.js"

    class Attributes(ComplexModel.Attributes):
        logged = False

    datatables = None

    title = Unicode(default='Neurons Screen')

    # meta properties
    description = Unicode(sub_name='content', default='Neurons Application')
    application_name = Unicode(sub_name='content',
                                                  default='Neurons Application')
    apple_mobile_webapp_title = Unicode(sub_name='content',
                                                          default='Neurons App')

    links = Array(Link, wrapped=False)
    styles = Array(CascadingStyleSheet, wrapped=False)
    scripts = Array(ScriptElement, wrapped=False)

    def __init__(self, ctx, *args, **kwargs):
        ComplexModel.__init__(self, *args, **kwargs)

        self._link_hrefs = set()
        self._have_clone = False
        self._have_jquery = False
        self._have_namespace = False
        self._have_urlencode = False
        self._have_type_checkers = False
        self._have_setup_datatables = False
        self._have_hide_empty_columns = False

        assert ctx.outprot_ctx.screen is None, \
                          "We are supposed to have only one screen per context."
        ctx.outprot_ctx.screen = self

    def append_script_href(self, what, type='text/javascript'):
        if self.scripts is None:
            self.scripts = []
        self.scripts.append(ScriptElement(href=what, type=type))

    def append_script(self, what, type='text/javascript', id=None):
        if self.scripts is None:
            self.scripts = []
        self.scripts.append(ScriptElement(what, type=type, id=id))

    def append_script_tree(self, tree, id=None, minify=False):
        if minify:
            mangler.mangle(tree, toplevel=True)
            ret = ECMAMinifier().visit(tree)
        else:
            ret = tree.to_ecma()

        self.append_script(ret, type='text/javascript', id=id)

    def append_style(self, what, media=None):
        if self.styles is None:
            self.styles = []
        self.styles.append(CascadingStyleSheet(what, media=media))

    def with_namespace(self):
        if not self._have_namespace:
            self.append_script("window.neurons = {}")

            self._have_namespace = True

        return self

    def with_jquery(self):
        if not self._have_jquery:
            self.append_script_href(self.JQUERY_URL)

            self._have_jquery = True

        return self

    def with_own_stylesheet(self, base='/assets/css/screen'):
        if self.links is None:
            self.links = []

        class_name = self.__class__.get_type_name()
        href = "%s/%s.css" % (base, class_name)

        if not (href in self._link_hrefs):
            self._link_hrefs.add(href)
            self.links.append(Link(rel="stylesheet", href=href))

        return self

    def with_setup_datatables(self):
        self.with_namespace()
        if not self._have_setup_datatables:
            self.append_script(SETUP_DATATABLES)

            self._have_setup_datatables = True
        return self

    def with_hide_empty_columns(self):
        self.with_namespace()
        if not self._have_hide_empty_columns:
            self.append_script(HIDE_EMPTY_COLUMNS)

            self._have_hide_empty_columns = True
        return self

    def with_datatables(self, data=None, hide_empty_columns=False):
        if data is None:
            data = self.datatables

        if data is None:
            data = {}

        self.with_jquery()

        if hide_empty_columns:
            self.with_hide_empty_columns()

        self.with_setup_datatables()

        for selector, data in data.items():
            retval = [
                "$(document).ready(function() {",
                    "neurons.setup_datatables(",
                        json.dumps(selector), ",",
                        json.dumps(data), ",",
                        json.dumps(hide_empty_columns),
                    ");",
                "});",
            ]

            self.append_script(''.join(retval))

        return self

    def with_type_checkers(self):
        if not self._have_type_checkers:
            self.with_namespace()

            self.append_script(TYPE_CHECKERS)

            self._have_type_checkers = True

        return self

    def with_clone(self):
        if not self._have_clone:
            self.with_namespace()

            self.append_script(CLONE)

            self._have_clone = True

        return self

    def with_urlencode(self):
        if not self._have_urlencode:
            self.with_namespace()

            self.append_script(URLENCODE)

            self._have_urlencode = True

        return self



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import os
import socket
import struct
import psutil

from spyne import rpc, UnsignedInteger16, Unicode
from neurons.base.service import TReaderService


def _gen_addr(base, num):
    # port numbers should never fall below 1024.
    mgmt_addr = base + num + 1024 * (num // 0xfc00)

    return (
        socket.inet_ntoa(struct.pack('!L', mgmt_addr >> 16)),
        mgmt_addr & 0xffff,
    )


MGMT_ADDR_BASE =   0x7f0100010400 # 127.1.0.1:1024
DOWSER_ADDR_BASE = 0x7f0101010400 # 127.1.1.1:1024


def get_dowser_address_for_pid(pid):
    """Computes dowser service address from process id.

    Returns a tuple containing the computed host as string and the port as int.
    """
    return _gen_addr(DOWSER_ADDR_BASE, pid)


def get_own_dowser_address():
    return get_dowser_address_for_pid(os.getpid())


def get_mgmt_address_for_pid(pid):
    """Computes management service address from process id.

    Returns a tuple containing the computed host as string and the port as int.
    """
    return _gen_addr(MGMT_ADDR_BASE, pid)


def get_mgmt_address_for_tcp_port(port):
    """Gets management service address from a tcp port.

    Returns a tuple containing the computed host as string and the port as int.
    """

    pid = None

    for conn in psutil.net_connections():
        if conn.status != 'LISTEN':
            continue

        h, p = conn.laddr
        if p == port:
            pid = conn.pid
            break

    if pid is not None:
        return get_mgmt_address_for_pid(pid)

    return None, None


def get_own_mgmt_address():
    return get_mgmt_address_for_pid(os.getpid())


class DaemonServices(TReaderService()):
    @rpc(Unicode, UnsignedInteger16)
    def unlisten(self, host, port):
        pass



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

from os.path import abspath
from spyne.util.resource import get_resource_path

from spyne import Application, File, Unicode
from spyne import Service, rpc
from spyne.protocol.html import HtmlCloth
from spyne.protocol.http import HttpRpc, HttpPattern

from .const import T_DOM_MODULE, T_SCREEN


class ComponentService(Service):
    @rpc(Unicode, Unicode,
        _patterns=[
            HttpPattern('neurons-<folder>/neurons-<file_name>.html', verb="GET"),
            HttpPattern('neurons-<file_name>.html', verb="GET"),
        ], _returns=File)
    def get_component(ctx, folder, file_name):
        ctx.out_protocol = HttpRpc()
        if folder is None:
            fn = "neurons-" + file_name
            fullpath = (get_resource_path('neurons.polymer.const.comp',
                                                     "{0}/{0}.html".format(fn)))
        else:
            fn = "neurons-%s/neurons-%s" % (folder, file_name)
            fullpath = (get_resource_path('neurons.polymer.const.comp',
                                                         "{0}.html".format(fn)))
        logger.debug("Returning component %s from path %s", fn, fullpath)
        return File.Value(path=fullpath, type="text/html")


def gen_component_app(config, prefix, classes, locale=None,
        gen_css_imports=False, api_read_url_prefix='', api_write_url_prefix=''):
    from neurons.polymer.protocol import PolymerForm
    from neurons.polymer.service import TComponentGeneratorService

    return \
        Application(
            [ComponentService] +
            [
                TComponentGeneratorService(
                    cls.customize(
                        prot=PolymerForm(mrpc_url_prefix='/'),
                    ),
                    prefix,
                    locale,
                    gen_css_imports,
                    api_read_url_prefix=api_read_url_prefix,
                    api_write_url_prefix=api_write_url_prefix,
                )
                for cls in classes
            ],
            tns='%s.comp' % config.name,
            in_protocol=HttpRpc(validator='soft'),
            out_protocol=HtmlCloth(cloth=T_DOM_MODULE, strip_comments=True,
                                                     doctype='<!DOCTYPE html>'),
            config=config,
        )


def gen_screen_services(prefix, classes, comp_prefix=''):
    from neurons.polymer.protocol import PolymerForm
    from neurons.polymer.service import TScreenGeneratorService

    return [
        TScreenGeneratorService(
            cls.customize(
                prot=PolymerForm(mrpc_url_prefix='/'),
            ),
            prefix, comp_prefix=comp_prefix,
        ) for cls in classes
    ]


def gen_screen_app(config, prefix, classes, comp_prefix=''):
    return \
        Application(
            gen_screen_services(prefix, classes, comp_prefix=comp_prefix),
            tns='%s.scr' % config.name,
            in_protocol=HttpRpc(validator='soft'),
            out_protocol=HtmlCloth(cloth=T_SCREEN, strip_comments=True,
                                                     doctype='<!DOCTYPE html>'),
            config=config,
        )
